/***********************************************************************
                          arsc.c  -  description
                          ----------------------
    begin                : Wed Apr 17 2001
    copyright            : (C) 2001-2004 by Georg P. Loczewski
    email                : gpl@lambda-bound.de
 **********************************************************************/

/***********************************************************************
*                                                                      *
* This program is free software; you can redistribute it and/or modify *
* it under the terms of the GNU General Public License as published by *
* the Free Software Foundation; either version 2 of the License, or    *
* (at your option) any later version.                                  *
*                                                                      *
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <assert.h>
#include <malloc.h>
#include "gc.h"
#include <string.h>
#include "arsclib.h"

extern jmp_buf _EXH;

ENV *globalEnv;

char *nameTable[MAX_NAMES];
NAME numNames;
NAME setName;
VALUE *setVal;
VALUE *falseValue, *trueValue;
VALUE *nil;

int false=0;
int true=1;


/*****************************************************************
 *                  ADT CONSTRUCTORS                             *
 *****************************************************************/

/* newVAL - make a value (evaluated expression)  */
VALUE *
newVAL (VTYPE t)
{
  VALUE *val;

  val = (VALUE *) GC_malloc (sizeof (VALUE));
  val->type = t;
  return val;
}


/* newLISTV - represent a value list as a value   */
VALUE *
newLISTV (VLIST *vl)
{
  VALUE *val;

  if (vl) {
     val = (VALUE *) GC_malloc (sizeof (VALUE));
     val->type = LISTV;
     val->val.lv.car = vl->head;
     val->val.lv.cdr = newLISTV(vl->tail);
     return val;
  }
  else {
     return NULL;
  }
}

/* newVLIST - represent a list as a value-list   */
VLIST *
newVLIST (VALUE *ls)
{
  VLIST *vl;
  
  if (ls) {
     vl = (VLIST *) GC_malloc (sizeof (VLIST));
     vl->head = ls->val.lv.car;
     vl->tail = newVLIST(ls->val.lv.cdr);
     return vl;
  }
  else {
     return NULL;
  }
}


/* newSYMV- make a symbol value */
VALUE *
newSYMV(NAME nm)
{
  VALUE *result;

  result = (VALUE *) GC_malloc (sizeof (VALUE));
  result->type = SYMV;
  result->val.sym = nm;
  return result;
}

/* newSTRV- make a string value */
VALUE *
newSTRV(NAME nm)
{
  VALUE *result;

  result = (VALUE *) GC_malloc (sizeof (VALUE));
  result->type = STRV;
  result->val.str = nm;
  return result;
}

/* newINTV- make an integer value */
VALUE *
newINTV (int i)
{
  VALUE *result;

  result = (VALUE *) GC_malloc (sizeof (VALUE));
  result->type = INTV;
  result->val.ival = i;
  return result;
}

/* newDBLV- make a floating point  value */
VALUE *
newDBLV (double d)
{
  VALUE *result;

  result = (VALUE *) GC_malloc (sizeof (VALUE));
  result->type = DBLV;
  result->val.dval = d;
  return result;
}

/* newCLAMV- make a c-lambda abstraction  */
VALUE *
newCLAMV (NLIST * vars, LFUN lfun, ENV *env)
{
  VALUE *cl;

  cl = (VALUE *) GC_malloc (sizeof (VALUE));
  cl->type = CLAMV;
  cl->val.clam.vars = vars;
  cl->val.clam.lfun = lfun;
  cl->val.clam.env = env;
  return cl;
}

/* ncons - make a list of names */
NLIST *
ncons (NAME nm, NLIST * nl)
{
  NLIST *newnl;

  newnl = (NLIST *) GC_malloc (sizeof (NLIST));
  newnl->head = nm;
  newnl->tail = nl;
  return newnl;
}


/* vcons - make a list of values */
VLIST *
vcons (VALUE * val, VLIST * vl)
{
  VLIST *newvl;

  newvl = (VLIST *) GC_malloc (sizeof (VLIST));
  newvl->head = val;
  newvl->tail = vl;
  return newvl;
}

/* cons - make a list of values */
VALUE *
cons (VALUE * hd, VALUE * tl)
{
  VALUE *newvl;

  newvl = (VALUE *) GC_malloc (sizeof (VALUE));
  newvl->type = LISTV;
  newvl->val.lv.car = hd;
  newvl->val.lv.cdr = tl;
  return newvl;
}

/* list2 - make a list of 2 values */
VALUE *list2 (VALUE *v1, VALUE * v2)
{
   return cons(v1, cons(v2, NULL));
}

/* list3 - make a list of 3 values */
VALUE *list3 (VALUE *v1, VALUE * v2, VALUE * v3)
{
   return cons(v1, list2(v2, v3));
}

/* list4 - make a list of 4 values */
VALUE *list4 (VALUE *v1, VALUE * v2, VALUE * v3, VALUE * v4)
{
   return cons(v1, list3(v2, v3, v4));
}

/* list5 - make a list of 5 values */
VALUE *list5 (VALUE *v1, VALUE * v2, VALUE * v3, VALUE * v4, 
	      VALUE * v5)
{
   return cons(v1, list4(v2, v3, v4, v5));
}

/* cons - make a list of 6 values */
VALUE *list6 (VALUE *v1, VALUE * v2, VALUE * v3, VALUE * v4,
              VALUE *v5, VALUE *v6)
{
   return cons(v1, list5(v2, v3, v4, v5, v6));
}

/* newENV - make an environment */
ENV *
newENV (NLIST * nl, VLIST * vl, ENV * env)
{
  ENV *newenv;

  newenv = (ENV *) GC_malloc (sizeof (ENV));
  newenv->vars = nl;
  newenv->values = vl;
  newenv->next = env;
  return newenv;
}


/*****************************************************************
 *                     NAMES OF VARIABLES                        *
 *****************************************************************/

/* initNames - place all pre-defined names into nameTable       */
void
initNames ()
{
  long i = 1;

  nameTable[i - 1] = "nil";
  i++;
  nameTable[i - 1] = "#f";
  i++;
  nameTable[i - 1] = "#t";
  numNames = i;
  falseValue = newVAL (SYMV);
  trueValue = newVAL (SYMV);
  nil = newVAL (SYMV);
  nil->val.sym = defName("nil");
  falseValue->val.sym = defName("#f");
  trueValue->val.sym = defName("#t");
}



/* defName - insert new name into nameTable                     */
NAME
defName (char *nm)
{
  long i = 1;
  int found = false;

  while (i <= numNames && !found) {
    if (!strcmp (nm, nameTable[i - 1]))
      found = true;
    else
      i++;
  }
  if (found)
    return i;
  if (i > MAX_NAMES) {
    printf ("No more room for names\n");
    longjmp (_EXH, 1);
  }
  numNames = i;
  nameTable[i - 1] = nm;
  return i;
}


/* prName - print name nm  */
void
prName (NAME nm)
{
  printf ("%s", nameTable[nm - 1]);
}


/* prString - print string nm (strings are handled like symbols ) */
void
prString (NAME nm)
{
  printf ("%s", nameTable[nm - 1]);
}

/* lengthNL - return length of NLIST nl      */
long
lengthNL (NLIST * nl)
{
  long i = 0;

  while (nl != NULL) {
    i++;
    nl = nl->tail;
  }
  return i;
}

/* prNL - print list of names */
void
prNL (NLIST * nl)
{
  long i = 0;

  putchar ('(');
  while (nl != NULL) {
    i++;
    prName (nl->head);
    nl = nl->tail;
    if (nl != NULL) putchar (' ');
  }
  printf (")\n");
}

/* prNLOLF - print list of names without LF*/
void
prNLOLF (NLIST * nl)
{
  long i = 0;

  putchar ('(');
  while (nl != NULL) {
    i++;
    prName (nl->head);
    nl = nl->tail;
    if (nl != NULL) putchar (' ');
  }
  printf (")");
}


/*****************************************************************
 *                     VARIABLES AND ENVIRONMENTS                *
 *****************************************************************/

/* emptyEnv - return an environment with no bindings             */
ENV *
emptyEnv ()
{
  return (newENV (NULL, NULL, NULL));
}


/* defVar - bind variable nm to value val in environment env      */
void
defVar (NAME nm, VALUE * val, ENV * env)
{
  env->vars = ncons (nm, env->vars);
  env->values = vcons (val, env->values);
}

/* setVar - set variable nm to value val in env                 */
void
setVar (NAME nm, VALUE * val, ENV * env)
{
  VLIST *vl;

  vl = bindingInEnv (nm, env);
  if (vl) {
     vl->head = val;
  }
  else {
     printf("variable not defined: ");
     prName(nm);
     printf("\n");
  }
}

/* getVar - return VAL bound to nm in env                        */
VALUE *
getVar (NAME nm, ENV * env)
{
  VLIST *vl;

  vl = bindingInEnv (nm, env);
  if (vl)
     return (vl->head);
  else
     return falseValue;
}

/*  extendEnv - extend environment env by binding vars to vals   */
ENV *
extendEnv (ENV * env, NLIST * vars, VLIST * vals)
{
  return (newENV (vars, vals, env));
}


/* bindingInFrame - look up nm in one frame                      */
VLIST *
bindingInFrame (NLIST * nl, VLIST * vl, NAME nm)
{
  int found = false;

  while (nl != NULL && !found) {
    if (nl->head == nm)
      found = true;
    else {
      nl = nl->tail;
      vl = vl->tail;
    }
  }
  if (lengthNL(nl) < lengthVL(vl)) {
     vl = vcons(newLISTV(vl), NULL);
  }
  return vl;
}


/* bindingInEnv - look up nm in env                                   */
VLIST *
bindingInEnv (NAME nm, ENV * env)
{
  VLIST *vl;

  do {
    vl = bindingInFrame (env->vars, env->values, nm);
    env = env->next;
  } while (vl == NULL && env != NULL);
  return vl;
}


/* isBound - check if nm is bound in env                         */
int
isBound (NAME nm, ENV * env)
{
  return (bindingInEnv (nm, env) != NULL);
}

/*****************************************************************
 *                     EVALUATION  AND VALUES                    *
 *****************************************************************/

/* sm - send message to objects */
VALUE *
sm (VALUE * obj, int imsg, VLIST *args)
{
   VALUE *result;

   result = apply (obj, mkmsg(imsg));
   result = apply (result, args);
}

VLIST *
mkmsg (int msg)
{
  return vcons (newINTV (msg), NULL);
}

/* apply - apply c-closure to arguments */
VALUE *
apply (VALUE * cl, VLIST * args)
{
  LFUN lfun;
  NLIST *names;
  ENV *savedenv, *newenv;

  savedenv = cl->val.clam.env;
  names = cl->val.clam.vars;
  lfun = cl->val.clam.lfun;
  if (lengthNL (names) > lengthVL (args)) {
    printf ("Wrong number of arguments to closure\n");
    printf ("names\n");
    prNL (names);
    printf ("arguments\n");
    prVL (args);
    longjmp (_EXH, 1);
  }
  newenv = extendEnv (savedenv, names, args);
  return (lfun(newenv));
}

/* prValue - print value  val                                */
void
prValue (VALUE * val)
{
  int i;
  
  if (val != NULL) {
     switch (val->type) {

     case INTV:
       printf ("%ld", val->val.ival);
       break;

     case SYMV:
       if (val != nil) prString (val->val.sym);
       break;

     case STRV:
       prString (val->val.str);
       break;

     case LISTV:
        i = 0;
        putchar ('(');
        while (val != NULL) {
          i++;
          prValue (val->val.lv.car);
          val = val->val.lv.cdr;
          if (val != NULL) putchar (' ');
        }
        printf (")\n");
       break;

     case CLAMV:
       printf ("<lambda");
       prNLOLF(val->val.clam.vars);
       printf (">");
       break;

     }
  }
}

/* prVal - print value  val with modified list processing  */
void
prVal (VALUE * val)
{
  int i;
  
  if (val != NULL) {
     switch (val->type) {

     case INTV:
       printf ("%ld", val->val.ival);
       break;

     case SYMV:
       if (val != nil) prString (val->val.sym);
       break;

     case STRV:
       prString (val->val.str);
       break;

     case LISTV:
        i = 0;
        putchar ('(');
        while (val != NULL) {
          i++;
          prVal (val->val.lv.car);
          val = val->val.lv.cdr;
          if (val != NULL) putchar (' ');
        }
        printf (")");
       break;

     case CLAMV:
       printf ("<lambda");
       prNLOLF(val->val.clam.vars);
       printf (">");
       break;

     }
  }
}

/* length - return length of common list */
long
length (VALUE * ls)
{
  long i = 0;

  if (ls->type == LISTV) {
     while (ls != NULL) {
       i++;
       ls = ls->val.lv.cdr;
     }
     return i;
  }
  else {
     printf("length applied to non-list!\n");
     return 0;
  }
}

/* lengthVL - return length of VLIST vl      */
long
lengthVL (VLIST * vl)
{
  long i = 0;

  while (vl != NULL) {
    i++;
    vl = vl->tail;
  }
  return i;
}

/* prVL - print list of values */
void
prVL (VLIST * vl)
{
  long i = 0;

  putchar ('(');
  while (vl != NULL) {
    i++;
    prValue (vl->head);
    putchar (' ');
    vl = vl->tail;
  }
  printf (")\n");
}

/* car - get first element in list */
VALUE * 
car(VALUE *ls) 
{
   if (ls->type == LISTV) {
         return ls->val.lv.car;
   }
   else {
      printf("car: list expected!\n");
      prValue(ls);
      return NULL;
   }
}
   
/* cdr - get tail in list */
VALUE * 
cdr(VALUE *ls) 
{
   if (ls->type == LISTV) {
         return ls->val.lv.cdr;
   }
   else {
      printf("cdr: list expected!\n");
      prValue(ls);
      return NULL;
   }
}
   
/* caar - get first element in first pair of list */
VALUE * 
caar(VALUE *ls) 
{
   if (ls->type == LISTV) {
         return car(ls->val.lv.car);
   }
   else {
      printf("caar: list expected!\n");
      prValue(ls);
      return NULL;
   }
}
   
/* cadr - get second element in list */
VALUE * 
cadr(VALUE *ls) 
{
   if (ls->type == LISTV) {
         return car(ls->val.lv.cdr);
   }
   else {
      printf("cadr: list expected!\n");
      prValue(ls);
      return NULL;
   }
}
   
/* nth - get nth element in list */
VALUE * 
nth(int n, VALUE *ls) 
{
   if (ls->type == LISTV) {
      if (n == 1) {
         return ls->val.lv.car;
      }
      else {
         return nth(n - 1, ls->val.lv.cdr);
      }
   }
   else {
      printf("nth: list expected!\n");
      prValue(ls);
      return NULL;
   }
}
   
VALUE *map(VALUE * proc, VALUE * ls) 
{
   assert(proc->type == CLAMV);
   if (ls == NULL){
      return NULL;
   }
   else {
      assert(ls->type == LISTV);
      return (cons (apply(proc, vcons(ls->val.lv.car,NULL)), 
                    map(proc, ls->val.lv.cdr)));
   }
}

int eqlist (VALUE * l1, VALUE * l2)
{
   if (nullp (l1) && nullp (l2)) {
      return true;
   }
   else {
      if (nullp (l1) || nullp (l2)) {
         return false;
      }
      else  {
         assert(l1->type == LISTV &&
	        l2->type == LISTV);
         return (equal (l1->val.lv.car, l2->val.lv.car) &&
                 eqlist (l1->val.lv.cdr, l2->val.lv.cdr));
      }
   }
}

int equal (VALUE * o1, VALUE * o2)
{
   if ((atomp (o1)) && (atomp (o2))){
      if (o1->type == o2->type) {
	 switch (o1->type) {
	    case SYMV:
		 return o1->val.sym == o2->val.sym;
	    case STRV:
		 return o1->val.str == o2->val.str;
	    case INTV:
		 return o1->val.ival == o2->val.ival;
	    default:
		 printf("equal: undefined data type!\n");
		 prValue(o1);
		 prValue(o2);
		 return false;
	 }
      }
      else {
         return false;
      }
   }
   else {
      if ((atomp (o1)) || (atomp (o2))){
         return false;
      }
      else {
         return eqlist (o1, o2);
      }
   }
}

VALUE *assoc(VALUE *x, VALUE *alist)
{
    if (nullp(alist)) {
       return falseValue;
    }
    else {
       if (equal(x,caar(alist))) {
          return car(alist);
       }
       else {
          return assoc(x,cdr(alist));
       }
    }
}

int  nullp(VALUE * li) 
{
   if (li == NULL)
      return 1;
   else
      return 0;
}

int  atomp(VALUE * obj) 
{
   if (obj->type == LISTV) {
      return 0;
   }
   else {
      return 1;
   }
}

