/**********************************************************************
                          arsc.h  -  description
                          ----------------------
    begin                : Wed Apr 17 2001
    copyright            : (C) 2001-2004 by Georg P. Loczewski
    email                : gpl@lambda-bound.de
**********************************************************************/

/***********************************************************************
*                                                                      *
* This program is free software; you can redistribute it and/or modify *
* it under the terms of the GNU General Public License as published by *
* the Free Software Foundation; either version 2 of the License, or    *
* (at your option) any later version.                                  *
*                                                                      *
***********************************************************************/
#ifndef _ARSC_H
#define _ARSC_H 
#define MAX_NAMES        300	/* Maximum number of different names */

typedef short NAME;		/* a NAME is an index in nameTable */

typedef enum
{
  INTV, DBLV, SYMV, STRV, LISTV, CLAMV
}
VTYPE;


typedef struct nlist
{
  NAME head;
  struct nlist *tail;
}
NLIST;

typedef struct vlist
{
  struct value *head;
  struct vlist *tail;
}
VLIST;

typedef struct env
{
  NLIST *vars;
  VLIST *values;
  struct env *next;
}
ENV;

typedef struct value  * (*LFUN) (ENV *env);

typedef struct clam
{
  struct nlist *vars;
  LFUN lfun;
  ENV  *env;
}
CLAM;

typedef struct value
{
  VTYPE type;
  union
  {
    long   ival;
    double dval;
    NAME   sym;
    NAME   str;
    struct {
       struct value  *car;
       struct value  *cdr;
    } lv;
    CLAM   clam;
  }
  val;
}
VALUE;


/*****************************************************************
 *                     ADT CONSTRUCTORS                          *
 *****************************************************************/
VALUE *newCLAMV (NLIST * vars, LFUN lfun, ENV *env);
VALUE *newVAL (VTYPE t);
VALUE *newLISTV (VLIST *vl);
VLIST *newVLIST (VALUE *ls);
VALUE *newSTRV (NAME nm);
VALUE *newSYMV (NAME nm);
VALUE *newINTV (int  i);
VALUE *newDBLV (double  d);
ENV *newENV (NLIST * nl, VLIST * vl, ENV * env);
NLIST *ncons (NAME nm, NLIST * nl);
VLIST *vcons (VALUE *val, VLIST * vl);
VALUE *cons (VALUE *hd, VALUE * tl);
VALUE *list2 (VALUE *v1, VALUE * v2);
VALUE *list3 (VALUE *v1, VALUE * v2, VALUE * v3);
VALUE *list4 (VALUE *v1, VALUE * v2, VALUE * v3, VALUE * v4);
VALUE *list5 (VALUE *v1, VALUE * v2, VALUE * v3, VALUE * v4,
              VALUE *v5);
VALUE *list6 (VALUE *v1, VALUE * v2, VALUE * v3, VALUE * v4,
              VALUE *v5, VALUE *v6);
/*****************************************************************
 *                     NAMES OF VARIABLES                        *
 *****************************************************************/
void initNames ();
NAME defName (char *nm);
void prName (NAME nm);
void prNL (NLIST * nl);
void prNLOLF (NLIST * nl);
long lengthNL (NLIST * nl);
void prString (NAME nm);
/*****************************************************************
 *                     VARIABLES AND ENVIRONMENTS                *
 *****************************************************************/
ENV *emptyEnv ();
void defVar (NAME nm, VALUE *val, ENV * env);
void setVar (NAME nm, VALUE *val, ENV * env);
VALUE *getVar (NAME nm, ENV * env);
ENV *extendEnv (ENV * env, NLIST * vars, VLIST * vals);
VLIST *bindingInFrame (NLIST * nl, VLIST * vl, NAME nm);
VLIST *bindingInEnv (NAME nm, ENV * env);
int isBound (NAME nm, ENV * env);
void initGlobalEnv ();
/*****************************************************************
 *                     VALUES                                    *
 *****************************************************************/
VLIST *mkmsg (int msg);
VALUE *sm (VALUE * obj, int imsg, VLIST *args);
VALUE *apply(VALUE * cl, VLIST * args);
void prValue (VALUE * val);
void prVal (VALUE * val);
void prVL (VLIST * vl);
long lengthVL (VLIST * vl);
long length (VALUE * ls);
VALUE *car (VALUE *val);
VALUE *cdr (VALUE *val);
VALUE *caar (VALUE *val);
VALUE *cadr (VALUE *val);
VALUE *nth (int n, VALUE *val);
VALUE *map(VALUE * proc, VALUE * ls);
int eqlist (VALUE * l1, VALUE * l2);
int equal (VALUE * o1, VALUE * o2);
VALUE *assoc(VALUE *x, VALUE *alist);
int  nullp(VALUE * li);
int  atomp(VALUE * obj);
/*****************************************************************
 *                     MAIN CONTROL AND INPUT                    *
 *****************************************************************/
int main (int argc, char *argv[]);
#endif
