;;;genfui.scm, A Dynamic Generic User Interface 
;;;            applied to XML-document processing
;;;Copyright (c) 2000 Georg P. Loczewski 
;
;Permission to copy this software, to redistribute it, and to use it
;for any purpose is granted, subject to the following restrictions and
;understandings.
;
;1.  Any copy made of this software must include this copyright notice
;in full.
;
;2.  I have made no warrantee or representation that the operation of
;this software will be error-free, and I am under no obligation to
;provide any services, by way of maintenance, update, or otherwise.
;
;3.  In conjunction with products arising from the use of this
;material, there shall be no use of my name in any advertising,
;promotional, or sales literature without prior written consent in
;each case.

(include "xmlutil.scm")
(load "xml2db.scm")

(define *welcome-msg* 
        "*** Structured Information Storage and Retrieval System ***")
(define main
  (lambda()
    (let ((genfui (make-genfui)))
      (display-alln "\t" *welcome-msg* )
      ((genfui 'menu-proc)))))

(define make-genfui
  (lambda()
    (letrec ((xmldb #f)
             (xml2db #f)
             (level 0)
             (selection #f)
             (doc-selection #f)

             (display-menu
               (lambda(menu)
                 (display-alln 
                   (map 
                     (lambda(x) 
                       (cond
                         ((equal? (nth 4 x) 'm)
                          (list (car x) 
                                (if (cadr x) 
                                    (list (cadr x)))  
                                    "-->"))
                         (else (list (car x) 
                                     (if (cadr x) 
                                         (list (cadr x))
                                          '())
                                     (nth 5 x)))))
                     menu))))

             (menu-proc
                (lambda()
                  (display-menu top-level)
                  (menu-proc-aux top-level)))

             (menu-proc-aux
               (lambda(menu)
                 (let* ((input (get-fn menu))
                        (fn-name (car input))
                        (args (cdr input))
                        (fn '?)
                        (menu-item '?))
                   (cond 
                     ((equal? fn-name 'help)
                      (let* ()
                        (display-menu menu)
                        (menu-proc-aux menu)))
                     ((equal? fn-name 'quit)
                      (set! level (- level 1))
                      'return)
                     (else
                       (set! menu-item (assoc fn-name menu))
                       (set! fn (nth 3 menu-item))
                       (cond 
                         ((equal? (nth 4 menu-item) 'f)
                          (if (not (in-domain? args menu-item))
                              (display-alln "argument check failed: " 
                                            args)
                              (display-alln (if (= 0 (length args))
                                                (fn)
                                                (apply fn args)) ))
                          (menu-proc-aux menu))
                         ((equal? (nth 4 menu-item) 'm)
                          (set! level (+ 1 level))
                          (display-menu (nth 5 menu-item))
                          (menu-proc-aux (nth 5 menu-item))
                          (display-menu menu)
                          (menu-proc-aux menu))
                         (else (display-alln "invalid type for menuitem: " 
                                         (nth 4 menu-item)))))))))
             
             (get-fn 
               (lambda(menu)
                 (display-all level ">>")
                 (let ((line (read)))
                   (if (not (pair? line))
                       (cond
                         ((symbol? line)
                          (cond
                            ((equal?  line  'q ) '(quit))
                            ((equal? line 'h) '(help))
                            (else
                              (let ((abk (abk? line menu)))
                                (if abk
                                    (list abk)
                                    '(help))))))
                         (else
                           '(help)))
                       (let ((fn-name (car line)))
                         (cond
                           ((or (equal? fn-name 'help)
                                (equal? fn-name 'quit))
                            line)
                           (else
                             (cond
                                ((equal? (assoc fn-name menu) #f)
                                 (display-alln (eval line))
                                 (get-fn menu))
                                (else
                                  line)))))))))
             
             (abk?
               (lambda (line menu) 
                 (if (null? menu) 
                     #f
                     (if (and (cadr (car menu)) 
                              (equal? line (cadr (car menu)))) 
                         (caar menu)
                         (abk? line (cdr menu) )))))
             
             (in-domain?
               (lambda(args menu-item)
                 (if (= (length args) (nth 5 menu-item))
                     (apply (nth 6 menu-item) args)
                     #f)))
             
             (db-create
               (lambda(name)
                 (if xmldb
                     (display-alln "database open: close first!")
                     (begin
                       (set! xmldb (make-xmldb name 'create))
                       (display-alln "database created: ok!")))))

             (db-open
               (lambda(name)
                 (if xmldb
                     (display-alln "database already open: close first!")
                     (if (file-exists? name)
                         (set! xmldb (make-xmldb name 'open))
                         (display-alln "database does not exist!")))))
             (db-save
               (lambda()
                 (if xmldb
                     (begin
                       ((xmldb 'db-save))
                       (set! xmldb #f)
                       (display-alln "database closed: ok!"))
                     (display-alln "database not open!"))))

             (db-select
               (lambda(scope props)
                 (if xmldb
                     (begin
                       (set! selection ((xmldb 'db-select) scope props))
                       (length selection))
                     (display-alln "database is not open!"))))
             
             (db-display
               (lambda()
                 (if xmldb
                     (if selection
                         ((xmldb 'db-export) selection "display")
                         (display-alln "no selection defined!"))
                     (display-alln "database is not open!"))))
             
             (db-import
               (lambda(file-name)
                 (if xmldb
                     ((xmldb 'db-import) file-name)
                     (display-alln "database is not open!"))))
             
             (db-export
               (lambda(file-name)
                 (if xmldb
                     (if selection
                         ((xmldb 'db-export) selection file-name)
                         (display-alln "no selection defined!"))
                     (display-alln "database is not open!"))))
             
             (xml-select
               (lambda(scope props)
                 (if xmldb
                     (begin
                       (set! scope (cons "xmldoc" scope))
                       (set! doc-selection ((xmldb 'db-select) scope 
                                                               props))
                       (set! selection doc-selection)
                       (length doc-selection))
                     (display-alln "database is not open!"))))
             
             (xml-display
               (lambda()
                 (db-display)))
             
             (xml-export
               (lambda(file-name)
                 (if xmldb
                     (if doc-selection
                         ((xmldb 'db-export) doc-selection file-name)
                         (display-alln "no selection defined!"))
                     (display-alln "database is not open!"))))
             
             (xml-import
               (lambda(file-name)
                 (if xmldb
                     ((xmldb 'db-import) file-name)
                     (display-alln "database is not open!"))))
             
             (xml-delete
               (lambda( )
                 (if xmldb
                     (if doc-selection
                       (begin
                         ((xmldb 'db-delete) doc-selection)
                         (set! doc-selection #f)
                         (set! selection #f))
                       (begin
                         (display-alln "no selection defined!")
                         #f))
                     (display-alln "database is not open!"))))
             (xml-update
               (lambda( )
                 (if xmldb
                     (if doc-selection
                       (begin
                         ((xmldb 'db-update) doc-selection)
                         (set! doc-selection #f)
                         (set! selection #f))
                       (begin
                         (display-alln "no selection defined!")
                         #f))
                     (display-alln "database is not open!"))))
             (top-level
               (list
                 (list 'Database 'd #f 'm 
                   (list
                     (list 'new #f db-create 'f 1 
                           (lambda(name) 'ok))
                     (list 'open #f db-open 'f 1 
                           (lambda(name) 'ok))
                     (list 'save 's db-save 'f 0 
                           (lambda() 'ok))))
                 (list 'XML-Document'x #f 'm 
                   (list
                     (list 'import #f xml-import 'f 1 
                           (lambda(name ) 'ok))
                     (list 'doc-select #f xml-select 'f 2 
                           (lambda(scope props ) 'ok))
                     (list 'display 'p xml-display 'f 0 
                           (lambda() 'ok))
                     (list 'delete 'd xml-delete 'f 0 
                           (lambda() 'ok))
                     (list 'update 'u xml-update 'f 0 
                           (lambda() 'ok))
                     (list 'export #f xml-export 'f 1 
                           (lambda(name ) 'ok))))
                 (list 'Node 'n #f 'm 
                   (list
                     (list 'select #f db-select 'f 2 
                           (lambda(scope props ) 'ok))
                     (list 'display 'p db-display 'f 0 
                           (lambda() 'ok))
                     (list 'export #f db-export 'f 1 
                           (lambda(name ) 'ok))))
                    ))
             (self
               (lambda(msg)
                 (cond
                   ((equal? msg 'menu-proc) menu-proc)
                   (else
                     (display "undefined operation: ")
                     (display msg)
                     (newline))))))
        self)))
