/***********************************************************************
                          main.c  -  description
                          ----------------------
    begin                : Mit Apr 17 22:22:30 CEST 2001
    copyright            : (C) 2001 by Georg P. Loczewski
    email                : gpl@lambda-bound.de
 **********************************************************************/

/***********************************************************************
*                                                                      *
* This program is free software; you can redistribute it and/or modify *
* it under the terms of the GNU General Public License as published by *
* the Free Software Foundation; either version 2 of the License, or    *
* (at your option) any later version.                                  *
*                                                                      *
***********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#include <setjmp.h>
#include <malloc.h>
#include "gc.h"
#include <string.h>
#include "arsclib.h"
#include "test.h"

#define  GET_BALANCE 1
#define  DEPOSIT 2
#define  WITHDRAW 3
#define  PRINT_BALANCE 4

extern ENV *globalEnv;

extern int true;

jmp_buf _EXH;

VALUE *
lf_make_account (ENV * env)
{
  VALUE *get_balance;
  VALUE *deposit;
  VALUE *withdraw;
  VALUE *print_balance;
  VALUE *self;

  get_balance = newCLAMV (NULL,lf_get_balance, env);
  deposit  = newCLAMV (ncons(defName("amount"), NULL),lf_deposit, env);
  withdraw = newCLAMV (ncons(defName("amount"),NULL),lf_withdraw, env);
  print_balance = newCLAMV (NULL,lf_print_balance, env);
  self = newCLAMV (ncons(defName("msg"),NULL), lf_self, env);
  defVar (defName ("get_balance"), get_balance, env);
  defVar (defName ("deposit"), deposit, env);
  defVar (defName ("withdraw"), withdraw, env);
  defVar (defName ("print_balance"), print_balance, env);
  return self;
}

VALUE *
lf_get_balance (ENV * env)
{
  return getVar (defName ("balance"), env);
}

VALUE *
lf_deposit (ENV * env)
{
  VALUE *balance;
  VALUE *amount;
  int ba, am;

  balance = getVar (defName ("balance"), env);
  amount = getVar (defName ("amount"), env);
  ba = balance->val.ival;
  am = amount->val.ival;
  balance->val.ival = ba + am;
  setVar (defName ("balance"), balance, env);
  return balance;
}

VALUE *
lf_withdraw (ENV * env)
{
  VALUE *balance;
  VALUE *amount;
  int ba, am;

  balance = getVar (defName ("balance"), env);
  amount = getVar (defName ("amount"), env);
  ba = balance->val.ival;
  am = amount->val.ival;
  balance->val.ival = ba - am;
  setVar (defName ("balance"), balance, env);
  return balance;
}

VALUE *
lf_print_balance (ENV * env)
{
  VALUE *balance;

  balance = getVar (defName ("balance"), env);
  printf ("balance: ");
  prValue (balance);
  printf ("\n");
  return balance;
}

VALUE *
lf_self (ENV * env)
{
  VALUE *ret;
  VALUE *vmsg;
  int msg;

  vmsg = getVar (defName ("msg"), env);
  msg = vmsg->val.ival;
  switch (msg) {
  case GET_BALANCE:
    ret = getVar (defName ("get_balance"), env);
    break;
  case DEPOSIT:
    ret = getVar (defName ("deposit"), env);
    break;
  case WITHDRAW:
    ret = getVar (defName ("withdraw"), env);
    break;
  case PRINT_BALANCE:
    ret = getVar (defName ("print_balance"), env);
    break;
  default:
    printf ("message not understood by object\n");
    longjmp (_EXH, 1);
  }
  return ret;
}

VLIST *
mkarg (int arg)
{
  return vcons (newINTV (arg), NULL);
}

VALUE * 
lf_konto (ENV * env)
{
  VALUE *make_account;
  VALUE *k1;
  VALUE *k2;
  VALUE *result;

  make_account = newCLAMV(ncons(defName("balance"),NULL),
		          lf_make_account, 
			  env);
  k1 = apply (make_account, mkarg(10));/*(define k1 (make-account ten)) */
  k2 = apply (make_account, mkarg(5));/*(define k2 (make-account five)) */

  result = sm (k1, PRINT_BALANCE, NULL);/* ((k1 msg-print)) */
  result = sm (k1, DEPOSIT, mkarg(10));	/* ((k1 msg-deposit) 10)*/
  result = sm (k1, PRINT_BALANCE, NULL);/* ((k1 msg-print)) */
  result = sm (k1, WITHDRAW, mkarg(4));	/* ((k1 msg-withdraw) 4) */
  result = sm (k1, PRINT_BALANCE, NULL);/* ((k1 msg-print)) */

  result = sm (k2, PRINT_BALANCE, NULL);/* ((k2 msg-print)) */
  result = sm (k2, DEPOSIT, mkarg(9));	/* ((k2 msg-deposit) 9) */
  result = sm (k2, PRINT_BALANCE, NULL);/* ((k2 msg-print)) */
  result = sm (k2, WITHDRAW, mkarg(7));	/* ((k2 msg-withdraw) 7)*/
  result = sm (k2, PRINT_BALANCE, NULL);/* ((k2 msg-print)) */
  return result;
}

void
ootest ()
{
  ENV *penv;
  VALUE *result;
  VALUE *konto;

  penv = emptyEnv ();
  konto = newCLAMV(NULL, lf_konto, penv);
  result = apply (konto, NULL);
}

VALUE *
mycar (ENV * env)
{
  VALUE *l;

  l = getVar (defName ("l"), env);
  if (l->type == LISTV) {
    return l->val.lv.car;
  }
  else {
    printf ("wrong argument type for car\n");
    longjmp (_EXH, 2);
  }
}

VALUE *
mykdr (ENV * env)
{
  VALUE *l;

  l = getVar (defName ("l"), env);
  if (l->type == LISTV) {
    return l->val.lv.cdr;
  }
  else {
    printf ("wrong argument type for cdr\n");
    longjmp (_EXH, 2);
  }
}

VALUE *
compose_aux (ENV * env)
{
  VALUE *f;
  VALUE *g;
  VALUE *x;
  VALUE *result;

  f = getVar (defName ("f"), env);
  g = getVar (defName ("g"), env);
  x = getVar (defName ("x"), env);
  result = apply (g, vcons (apply (f, vcons (x, NULL)), NULL));
  return result;
}

VALUE *
compose (ENV * env)
{
  return newCLAMV(ncons (defName ("x"), NULL), 
		compose_aux,
		env);
}

void
futest ()
{
  ENV *penv;
  VALUE *result;
  VALUE *cadr;
  VALUE *clam_mycar;
  VALUE *clam_mykdr;
  VALUE *clam_compose;
  VLIST *vl;

  penv = emptyEnv ();
  vl =
    vcons(cons(newINTV (2),
	       cons(newINTV(3),
		    cons(newINTV(4),
			 cons(newINTV(6),
			      NULL)))), NULL);
  clam_mycar = newCLAMV (ncons(defName ("l"),NULL),mycar, penv);
  clam_mykdr = newCLAMV (ncons(defName ("l"),NULL),mykdr, penv);
  clam_compose = newCLAMV (ncons(defName("f"), ncons(defName("g"),
                                                     NULL)),
		         compose, 
			 penv);
  cadr = apply (clam_compose, vcons(clam_mykdr,vcons(clam_mycar, NULL)));
  result = apply (cadr, vl);
  prValue (result);
  printf ("\n");
}

  
 
int
main (int argc, char *argv[])
{
  int ret;

  //malloc_stats ();
  initNames ();

  if (argc > 1) {
    switch (argv[1][1]) {
    case 'o':
      ret = setjmp (_EXH);
      if (ret == 1 || ret == 2)
	goto _LQUIT;
      printf ("test of object oriented  programming style:\n");
      ootest ();
      longjmp (_EXH, 2);
    case 'f':
      ret = setjmp (_EXH);
      if (ret == 1 || ret == 2)
	goto _LQUIT;
      printf ("test of functional programming style:\n");
      futest ();
      longjmp (_EXH, 2);
    default:
      printf ("undefined option: %s\n", argv[1]);
      longjmp (_EXH, 2);
    }
  }
  else {
     ret = setjmp (_EXH);
     goto _LQUIT;
  }
_LQUIT:
  //malloc_stats ();
  return EXIT_SUCCESS;
}
