/***********************************************************************
                          valparse.c  -  description
                             -------------------
    begin                : Wed Mar 21 2001
    copyright            : (C) 2001-2004 by Georg P. Loczewski
    email                : gpl@lambda-bound.de
************************************************************************/

/************************************************************************
*                                                                       *
* This program is free software; you can redistribute it and/or modify  *
* it under the terms of the GNU General Public License as published by  *
* the Free Software Foundation; either version 2 of the License, or     *
* (at your option) any later version.                                   *
*                                                                       *
************************************************************************/
# include <stdio.h>
# include <string.h>
# include <ctype.h>

# include "valparse.h"


  static char buffer[1000];
  static char *current;		// current buffer pointer
  static FILE *ins;		// input stream

  static void printPrimaryPrompt ();
  static void printSecondaryPrompt ();
  static void primeBuffer ();
  static int isSeparator (int);
  static void skipSpaces ();
  static void skipNewlines ();

  static VALUE * readValue ();
  static VALUE * readSymbol ();
  static VALUE * readString ();
  static VALUE * readInt ();
  static VALUE * readList ();

void valparsinit (FILE * is)
{
  if (is == NULL) {
     ins = stdin;
  }
  else {
     ins = is;
  }
}

static void
printPrimaryPrompt ()
{
  printf ("\ninput-> ");
  fflush (stdout);
}

static void
printSecondaryPrompt ()
{
  printf ("> ");
  fflush (stdout);
}


static void
primeBuffer ()
{
  char *p;

  p = fgets (buffer, 1000, ins);
  if (p == NULL) {		// eof return by fgets
    if (ins == stdin) {
      strcpy (buffer, "quit ");
    }
    else {
      strcpy (buffer, "eof ");
    }
    p = buffer;
  }

  // initialize the current pointer
  *(p + strlen (p) - 1) = '\0';	// erase the newline character
  current = buffer;
  skipSpaces ();
}

static int
isSeparator (int c)
{
  switch (c) {
  case ' ':
  case '\t':
  case '\n':
  case EOF:
  case '\0':
  case '\'':
  case ';':
  case ')':
  case '(':
    return 1;
  }
  return 0;
}

static void
skipSpaces ()
{
  while ((*current == ' ') || (*current == '\t'))
    current++;
  if (*current == ';') {	// ignore comments
    while (*current)
      current++;
  }
}

static void
skipNewlines ()
{
  skipSpaces ();
  while (*current == '\0') {	// end of line
    printSecondaryPrompt ();	// print prompt as long as buffer is
    primeBuffer ();		// empty
  }
}

VALUE *
readWithPrompt ()
{
  VALUE * obj;

  do {
    printPrimaryPrompt ();
    primeBuffer ();
  } while (*current == '\0');

  obj = readValue ();

  skipSpaces ();
  if (*current) {
    error ("unexpected characters at end of line:", current);
  }
  return obj;
}

VALUE *
readWithoutPrompt ()
{
  VALUE * obj;

  do {
    primeBuffer ();
  } while (*current == '\0');

  obj = readValue ();

  // make sure we are at and of line
  skipSpaces ();
  if (*current) {
    error ("unexpected characters at end of line:", current);
  }
  return obj;
}

static VALUE *
readValue ()
{
  if (*current == '(') {
    current++;
    return readList ();
  }
  else if (*current == '"') {
    current++;
    return readString ();
  }
  else if (isdigit(*current)) {
    return readInt ();
  }
  else {
    return readSymbol ();
  }
}

static VALUE *
readList ()
{
   VALUE *hd;
   VALUE *tl;

   skipNewlines ();

   if (*current == ')') {
     current++;
     return NULL;
   }
   hd =  readValue ();
   tl =  readList ();
   return cons(hd, tl);
}

static VALUE *
readSymbol ()
{
  char *symbuf, *p;
  int len;

  symbuf = (char *) GC_malloc(sizeof(char) * 41);
  for (p = symbuf,len = 0; 
       len < 40 && !isSeparator (*current); 
       len++, *p++ = *current++);
  *p = '\0';
  return newSYMV(defName(symbuf));
}

static VALUE *
readString ()
{
  char *strbuf, *p;
  int len;

  strbuf = (char *) GC_malloc(sizeof(char) * 81);
  for (p = strbuf,len = 0; 
       len < 40 && !(*current == '"'); 
       len++, *p++ = *current++);
  current++;
  *p = '\0';
  return newSYMV(defName(strbuf));
}

static VALUE *
readInt ()
{
  char *ibuf, *p;
  int len;

  ibuf = (char *) GC_malloc(sizeof(char) * 21);
  for (p = ibuf,len = 0; 
       len < 20 && isdigit (*current); 
       len++, *p++ = *current++);
  *p = '\0';
  return newINTV(atoi(ibuf));
}

